#ifndef zc030x_jpeg_h
#define zc030x_jpeg_h

#include <asm/types.h>

typedef struct
{
//  u8 A;
  u8 B;
  u8 G;
  u8 R;
} Pixel24;

typedef struct
{
	u8 B;
	u8 G;
	u8 R;
	u8 x;
} Pixel32;

typedef struct
{                  
  int           Index;          // Index of this minimum code entry
  long          CodeValue;      // And its code value
} MinimumCodeEntry;


typedef struct 
{
  MinimumCodeEntry * pMinCodeTable;     // Minimum code table (given the huffman tree) and a length, as index
  long *             pMaxCodeTable;     // Maximum code table (given the huffman tree) and a length, as index
  unsigned char *    pCodeCatTable;     // The code category (bit length) table
  unsigned short *   pCodeTable;        // The code itself table
} HuffmanTable;


typedef unsigned char  byte;
typedef unsigned short word;

// Define functions
// Construct the huffman tables
void ConstructHuffmanTables(void);
// Get n bit from the huffman stream
int GetNBits(int NumberOfBits, byte **); 
// Get code from this stream given the huffman table
int GetCode(HuffmanTable * pTable, byte ** Buf); 
// Decode a code
int DecodeCode(int Code, int NumberOfBits); 

// Invert DCT
//TBRM: static void IDCT (short Data[]);
// InverDCT caller plus dequantifier
int InvertDCT(short vector[], byte quant[]);
// InitDecoder
void InitDecoder(void);
// decode frame
//int zc030x_decodeFrame(struct zc030x_frame *frame,  int format)

#define DCTSIZE		8	/* The basic DCT block is 8x8 samples */
#define DCTSIZE2	64	/* DCTSIZE squared; # of elements in a block */
#define ONE 1L			/* remove L if long > 32 bits */

#ifdef RIGHT_SHIFT_IS_UNSIGNED
#define LG2_DCT_SCALE 15
#define RIGHT_SHIFT(_x,_shft)   ((((_x) + (ONE << 30)) >> (_shft)) - (ONE << (30 - (_shft))))
#else
#define LG2_DCT_SCALE 16
#define RIGHT_SHIFT(_x,_shft)   ((_x) >> (_shft))
#endif

#define DCT_SCALE (ONE << LG2_DCT_SCALE)

#define LG2_OVERSCALE 2
#define OVERSCALE (ONE << LG2_OVERSCALE)

#define FIX(x)  ((int) ((x) * DCT_SCALE + 0.5))
#define FIXO(x)  ((int) ((x) * DCT_SCALE / OVERSCALE + 0.5))
#define UNFIX(x)   RIGHT_SHIFT((x) + (ONE << (LG2_DCT_SCALE-1)), LG2_DCT_SCALE)
#define UNFIXH(x)  RIGHT_SHIFT((x) + (ONE << LG2_DCT_SCALE), LG2_DCT_SCALE+1)
#define UNFIXO(x)  RIGHT_SHIFT((x) + (ONE << (LG2_DCT_SCALE-1-LG2_OVERSCALE)), LG2_DCT_SCALE-LG2_OVERSCALE)
#define OVERSH(x)   ((x) << LG2_OVERSCALE)

#define SIN_1_4 FIX(0.7071067811856476)
#define COS_1_4 SIN_1_4

#define SIN_1_8 FIX(0.3826834323650898)
#define COS_1_8 FIX(0.9238795325112870)
#define SIN_3_8 COS_1_8
#define COS_3_8 SIN_1_8

#define SIN_1_16 FIX(0.1950903220161282)
#define COS_1_16 FIX(0.9807852804032300)
#define SIN_7_16 COS_1_16
#define COS_7_16 SIN_1_16

#define SIN_3_16 FIX(0.5555702330196022)
#define COS_3_16 FIX(0.8314696123025450)
#define SIN_5_16 COS_3_16
#define COS_5_16 SIN_3_16

#define OSIN_1_4 FIXO(0.707106781185647)
#define OCOS_1_4 OSIN_1_4

#define OSIN_1_8 FIXO(0.3826834323650898)
#define OCOS_1_8 FIXO(0.9238795325112870)
#define OSIN_3_8 OCOS_1_8
#define OCOS_3_8 OSIN_1_8

#define OSIN_1_16 FIXO(0.1950903220161282)
#define OCOS_1_16 FIXO(0.9807852804032300)
#define OSIN_7_16 OCOS_1_16
#define OCOS_7_16 OSIN_1_16

#define OSIN_3_16 FIXO(0.5555702330196022)
#define OCOS_3_16 FIXO(0.8314696123025450)
#define OSIN_5_16 OCOS_3_16
#define OCOS_5_16 OSIN_3_16

/*typedef unsigned int   u32;
typedef unsigned short u16;
typedef unsigned char  u8;
typedef signed int     s32;
typedef signed short   s16;
typedef signed char    s8;*/

#endif

