/*
 * zc030x driver ( experimental )
 *
 * Authors: 
 * 	       Martin Braun (martinbraun at gmx.net)
 *             Arnaud Patard (thertp at users.sourceforge.net)
 *             Andre Freitas (linux at uninet.com.br)
 *             Cyril Russo ( )
 *             Andy Birkett (andy at nobugs.org)
 *             Tommy (illth at gmx.de)
 * 
 *
 * code is mainly based on: 
 *         
 *             SPCA50x webcam driver 
 *             CPiA webcam driver
 *             Phillips webcam drivers (pwc) 
 *             USB Skeleton Driver 0.7
 *
 * see ChangeLog for code changes...
 * 
 */


#include <linux/config.h>
#include <linux/version.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#define MODVERSIONS
#endif
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#endif

/*definetely needed */
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/usb.h>
 
/*questionably needed */
#include <linux/fs.h>
#include <linux/vmalloc.h>
#include <linux/string.h>
#include <linux/mman.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
#include <linux/wrapper.h>
#endif
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/poll.h>
#include <linux/slab.h>
#include <linux/fcntl.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/smp_lock.h>

//#include "zc030x_stream.h"
//replace with
#include "zc030x_jpeg.h"

#include "zc030x_main.h"
#include "zc030x_proc.h"
//#include "zc030x_v4l.h"
#include "zc030x_compat.h"
/* old sensor init headers - TODO: to be replaced by new ones below */
#include "webcamnx.h"
#include "wcam300a.h"
#include "webcamnxpro.h"
#include "webcamnotebook.h"
#include "benqcam2.h"
#include "zstar.h"
#include "webcammobile.h"

/* sensor init headers*/
#include "./sensors/CS2102.h"
#include "./sensors/CS2103.h"
#include "./sensors/HDCS1020.h"
#include "./sensors/HDCS2020.h"
#include "./sensors/HV7121B.h"
#include "./sensors/HV7131B.h"
#include "./sensors/ICM102A.h"
#include "./sensors/ICM105A.h"
#include "./sensors/OV7620.h"
#include "./sensors/OVCIF.h"
#include "./sensors/PAS106B.h"
#include "./sensors/PAS202B.h"
#include "./sensors/PB0111.h"
#include "./sensors/PB0330.h"
#include "./sensors/TAS5130C.h"
#include "./sensors/TAS5130C2.h"
#include "./sensors/TAS5110B.h"

// for 2.6 compatibility
// nasty hack ;)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#define USB_ISO_ASAP URB_ISO_ASAP
#endif 

/* Module paramaters */
static int debug = -1;

MODULE_PARM (debug, "i");
MODULE_PARM_DESC (debug, "Debug level: 0=no debug, 1=init/detection, 2=warning, 3=config/control, 4=function call, 5=max");
MODULE_AUTHOR("Martin Braun <martinbraun at gmx.net>, Arnaud Patard <thertp at users.sourceforge.net>, Andre Freitas <linux at uninet.com.br>, Cyril Russo<>, Andrew Birkett <andy@nobugs.org>");
MODULE_DESCRIPTION("ZC030X USB Camera Driver");
MODULE_LICENSE("GPL");

/* table of devices that work with this driver */
static struct usb_device_id zc030x_table[] = {
      {USB_DEVICE (USB_W300A_VENDOR_ID, USB_W300A_PRODUCT_ID)},
      {USB_DEVICE (USB_WEBCAMNXPRO_VENDOR_ID, USB_WEBCAMNXPRO_PRODUCT_ID)},
      {USB_DEVICE (USB_WEBCAMNX_VENDOR_ID, USB_WEBCAMNX_PRODUCT_ID)},
      {USB_DEVICE (USB_GENIUSV2_VENDOR_ID, USB_GENIUSV2_PRODUCT_ID)},
      {USB_DEVICE (USB_BENQCAM_VENDOR_ID, USB_BENQCAM_PRODUCT_ID)},
      {USB_DEVICE (USB_ZSTARCORP_VENDOR_ID, USB_ZSTARCORP_PRODUCT_ID)},
      {USB_DEVICE (USB_WEBCAMMOBILE_VENDOR_ID, USB_WEBCAMMOBILE_PRODUCT_ID)},
      {USB_DEVICE (USB_WEBCAMNOTEBOOK_VENDOR_ID, USB_WEBCAMNOTEBOOK_PRODUCT_ID)},
      {}                            /* Terminating entry */
};

MODULE_DEVICE_TABLE (usb, zc030x_table);

/* symbolic constants for type_Id field in device struct */
enum {
	W300A,
	WEBCAMNXPRO,
	WEBCAMNX,
	GENIUSV2,
	BENQCAM,
    ZSTARCORP,
	WEBCAMMOBILE,
	WEBCAMNOTEBOOK,
	UNKNOWN
} supported_cams;

/* list of camera names (in order with zc030x_table[] */
static struct cam_list clist[]={
       {W300A, "Mustek Webcam 300A"},
       {WEBCAMNXPRO, "Creative Webcam NX Pro"},
       {WEBCAMNX, "Creative Webcam NX"},
       {GENIUSV2,"Genius VideoCam Web V2"},
       {BENQCAM, "BenQ WebCam"}, 
       {ZSTARCORP, "Z-Star Corp PC Camera"}, 
       {WEBCAMMOBILE, "Creative Webcam Mobile"}, 
       {WEBCAMNOTEBOOK,"Creatove Webcam Notebook"},
       {-1, NULL}
};



/* Dirty but quick definition for same structures */ 
#define DeclareSensor(X)    { X, #X, 0, 0, 640, 480, &sig_##X }
struct sensor_list slist[] = 
{
    DeclareSensor(CS2102),
    DeclareSensor(CS2103),
 	DeclareSensor(HDCS1020),
	DeclareSensor(HDCS2020),
	DeclareSensor(ICM102A),
	DeclareSensor(ICM105A),
	DeclareSensor(HV7121B),
	DeclareSensor(HV7131B),
	DeclareSensor(PB0111),
	DeclareSensor(PB0330),
	DeclareSensor(PAS106B),
	DeclareSensor(PAS202B),
	DeclareSensor(OVCIF),
	DeclareSensor(OV7620),
	DeclareSensor(TAS5110B),
	DeclareSensor(TAS5130C),
        {-1, NULL}
};


/* device specific default alternate configuration */
/* find a suitable alternate config using usbview */
/* TODO: might autodetect this in future */
int default_alt_setting[]={
        7, /* Mustek Webcam 300A */
        1, /* Creative Webcam NX Pro */
	1, /* Creative Webcam NX */
	7, /* Genius VideoCam Web V2 */
	1, /* BenQCam ???? */
    1,
};

int FrameWidth = 320;
int FrameHeight = 232;
/* Value for Webcam Mobile
int FrameHeight = 240;
*/

/* struct for the tasklets */
struct zc030x_frame *taskletframe;
void outpict_do_tasklet (unsigned long ptr); 
DECLARE_TASKLET (outpicttasklet,outpict_do_tasklet,(unsigned long)&taskletframe);

/* v4l device functions */
static ssize_t zc030x_read (struct file *file, char *buffer, size_t count,
			    loff_t * ppos);
static int zc030x_ioctl (struct inode *inode, struct file *file,
			 unsigned int cmd, unsigned long arg);
static int zc030x_mmap(struct file *file, struct vm_area_struct *vma);
static int zc030x_open (struct inode *inode, struct file *file);
static int zc030x_release (struct inode *inode, struct file *file);

/* usb device functions */
/* TODO: change to be used with kernel 2.6 */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) 
static void *zc030x_probe (struct usb_device *dev, unsigned int ifnum,
		const struct usb_device_id *id);
static void zc030x_disconnect (struct usb_device *dev, void *ptr);
#else
static int zc030x_probe (struct usb_interface *interface, const struct usb_device_id *id);
static void zc030x_disconnect (struct usb_interface *interface);
#endif

/* lock to protect the minor_table structure */
static DECLARE_MUTEX (disconnect_sem);

/* usb specific object needed to register this driver with the usb subsystem */
static struct usb_driver zc030x_driver = {
    .owner =        THIS_MODULE,
    .name =	    "zc030x",
    .probe =	    zc030x_probe,
    .disconnect =   zc030x_disconnect,
    .id_table =     zc030x_table,
};

/* file operations supported by the v4l video device */
static struct file_operations zc030x_fops = {
    .owner =	    THIS_MODULE,
    .open =	    zc030x_open,
    .release =	    zc030x_release,
    .read =	    zc030x_read,
    .ioctl =	    zc030x_ioctl,
    .mmap =         zc030x_mmap, //NULL, //TODO: zc030x_mmap, - needs to be merged from older driver version (zc0302_v4l.c) 
    .llseek =       no_llseek,
};

#define VID_HARDWARE_ZC030X   37

static struct video_device zc030x_template = {
    .owner =         THIS_MODULE,
    .name =          "ZC030X Webcam",
    .type =          VID_TYPE_CAPTURE,
    .hardware =      VID_HARDWARE_ZC030X,
    .fops =          &zc030x_fops,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) 
    .initialize =    NULL,                   // initialize
#endif
    .minor=          -1                      // minor 
};



/**********************************************************************
 *
 * Memory management
 *
 * another shameless copy......
 *
 * This is a shameless copy from the USB-cpia driver (linux kernel
 * version 2.3.29 or so, I have no idea what this code actually does ;).
 * Actually it seems to be a copy of a shameless copy of the bttv-driver.
 * Or that is a copy of a shameless copy of ... (To the powers: is there
 * no generic kernel-function to do this sort of stuff?)
 *
 * Yes, it was a shameless copy from the bttv-driver. IIRC, Alan says
 * there will be one, but apparentely not yet -jerdfelt
 *
 * So I copied it again for the ov511 driver -claudio
 * And again for the spca50x driver -jcrisp
 **********************************************************************/


static void *rvmalloc(unsigned long size)
{
	void *mem;
	unsigned long adr;

	size = PAGE_ALIGN(size);
	mem = vmalloc_32(size);
	if (!mem)
	return NULL;

	memset(mem, 0, size); /* Clear the ram out, no junk to the user */
	adr = (unsigned long) mem;
	while ((long) size > 0) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 68)
		SetPageReserved(vmalloc_to_page((void *)adr));
#else
		mem_map_reserve(vmalloc_to_page((void *)adr));
#endif
		adr += PAGE_SIZE;
		size -= PAGE_SIZE;
	}
	
	return mem;
}

static void
rvfree(void *mem, unsigned long size)
{
	unsigned long adr;

	if (!mem)
		return;

	adr = (unsigned long) mem;
	while ((long) size > 0) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 68)
		ClearPageReserved(vmalloc_to_page((void *)adr));
#else
		mem_map_unreserve(vmalloc_to_page((void *)adr));
#endif
		adr += PAGE_SIZE;
		size -= PAGE_SIZE;
	}
	vfree(mem);
}

/* Given PGD from the address space's page table, return the kernel
 * virtual mapping of the physical memory mapped at ADR.
 */
static inline unsigned long uvirt_to_kva(pgd_t *pgd, unsigned long adr)
{
	unsigned long ret = 0UL;
	pmd_t *pmd;
	pte_t *ptep, pte;

	if (!pgd_none(*pgd)) {
		pmd = pmd_offset(pgd, adr);
		if (!pmd_none(*pmd)) {
			/* WOLT kernels appear to have changed pte_offset to pte_offset_kernel */
#ifdef pte_offset		
			ptep = pte_offset(pmd, adr);
#else /* pte_offset */
			ptep = pte_offset_kernel(pmd, adr);
#endif /* pte_offset */
			pte = *ptep;
			if (pte_present(pte)) {
				ret = (unsigned long) page_address(pte_page(pte));
				ret |= (adr & (PAGE_SIZE - 1));
			}
		}
	}

	return ret;
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
inline unsigned long kvirt_to_pa(unsigned long adr)
{
	unsigned long kva, ret;

	kva = (unsigned long) page_address(vmalloc_to_page((void *) adr));
	kva |= adr & (PAGE_SIZE - 1);
	ret = __pa(kva);
	return ret;
}
#else 
inline unsigned long kvirt_to_pa(unsigned long adr)
{
	unsigned long va, kva, ret;

	va = VMALLOC_VMADDR(adr);
	kva = uvirt_to_kva(pgd_offset_k(va), va);
	ret = __pa(kva);
	return ret;
}
#endif


/***************************************************************************
 *
 *
 *  ISOC transfer initialization & frame handling
 *  
 *
 **************************************************************************/

/**
  * zc030x_next_frame 
  * 
  * returns next free framebuffer to grab into
  * 
  */

static struct zc030x_frame *zc030x_next_frame(struct usb_zc030x *dev,unsigned char *cdata)
{
   int iFrameNext;
   struct zc030x_frame *frame=NULL;
   if( dev->frame[dev->curframe].grabstate == FRAME_ERROR) 
   {
     /* recycle the bad one */
     frame=&dev->frame[dev->curframe];
   } 
  else 
  {
      /* Cycle through the frame buffer looking for a free frame to overwrite */
      iFrameNext = (dev->curframe + 1) % ZC030X_NUMFRAMES;
      while(frame==NULL && iFrameNext!=(dev->curframe))
      {
	if (dev->frame[iFrameNext].grabstate == FRAME_READY  ||
	    dev->frame[iFrameNext].grabstate == FRAME_UNUSED ||
	    dev->frame[iFrameNext].grabstate == FRAME_ERROR)
	{
	   dev->curframe = iFrameNext;
	   frame=&dev->frame[iFrameNext];
	   break;
	}
	else
	{
	   iFrameNext=(iFrameNext+1)%ZC030X_NUMFRAMES;
	}
      }
   }
  if(frame==NULL)
  {
    PDEBUG(4,"Can't find a free frame to grab into...using next. ");
    PDEBUG(4,"This is caused by the application not reading fast enough.");
    dev->curframe=(dev->curframe + 1) % ZC030X_NUMFRAMES;
    frame=&dev->frame[dev->curframe];
  }
  
  frame->grabstate=FRAME_GRABBING;

  /* Reset some per-frame variables */
  frame->highwater=frame->data;
  frame->scanstate = STATE_LINES;
  frame->last_packet=-1;

  return frame;
  
}

/**
  * outpic_do_tasklet
  *
  * tasklet that does the decoding
  * 
  * TODO: I think more check needed and some spinlock code
  */
void outpict_do_tasklet (unsigned long ptr)
{       
  int err;
  int format=0;
 
  memcpy(taskletframe->tmpbuffer, taskletframe->data,
  	      taskletframe->highwater - taskletframe->data);

  PDEBUG(2, "Tasklet ask spcadecoder hdrwidth %d hdrheight %d", 
	      taskletframe->hdrwidth,taskletframe->hdrheight);
 
  err = zc030x_decodeFrame( taskletframe, format);
  PDEBUG(2, "zc030x_decodeFrame returned with code=%d",err);
 
  if (err != 0)
  {
    PDEBUG(2, "frame decoder failed (%d)", err);
  }
 
  taskletframe->grabstate = FRAME_DONE;
  if (waitqueue_active(&taskletframe->wq));
  wake_up_interruptible(&taskletframe->wq);
}

/**
  * zc030x_move_data 
  *
  * moves the data received in the iso pipe into framebuffers
  * and schedules decoder tasklet
  * 
  * TODO: I think more check needed and some spinlock code
  */
static int zc030x_move_data (struct usb_zc030x *dev, struct urb *urb)
{
     int i;
     int datalength;
     int st;
     int sequenceNumber;
     int totlen = 0; //bytes copied
     unsigned char *cData;  //Pointer to buffer where we find next packet
     unsigned char *pData;  //Pointer to buffer where we find next packet
     struct zc030x_frame *frame; //Pointer to frame data

     // Lock the buffers to ensure that they don't vanish while we are 
     // writing to them !
     if(down_trylock(&dev->buf_lock)) return -ENOBUFS;

     // Verify the buffers still exist
     if(!dev->fbuf)
     {
        up(&dev->buf_lock);
        return -ENOBUFS;
     }
     
     for (i = 0; i < urb->number_of_packets; i++)
     {
       datalength = urb->iso_frame_desc[i].actual_length;
       st = urb->iso_frame_desc[i].status;
       sequenceNumber=0;
       
       /* Check for zero length block or no selected frame buffer 
          if yes, skip the packet....                             */
       if (!datalength || dev->curframe == -1)
       continue;
      
       /* ...else packet data available -> cdata points to packet data! */
       cData = ((unsigned char *) urb->transfer_buffer) +
	                          urb->iso_frame_desc[i].offset;
      
       if (st)
       PDEBUG(2, "data error: [%d] len=%d, status=%d", i, datalength, st);

       /* get pointer to curret framebuffer */
       frame = &dev->frame[dev->curframe];

       /* TODO: frame dropping, frame error handling */
       /*       Andrew got an idea about this... (interrupt messages?) */
       
       /* Check if we find a SOI... */
       if( ((*((int *)cData)) & 0xFFFF) == 0xD8FF )
       {
        
	    PDEBUG(5,"frame compressed size: %d",(int)(frame->highwater-frame->data));
            /* Schedule tasklet to decode this frame */
	    //FIXME: shitty quickhack
	    if(frame->seq > 1)
	    {
	      taskletframe = frame;
	      taskletframe->format = dev->videoformat;
	      tasklet_schedule(&outpicttasklet);
	
	      if (waitqueue_active(&frame->wq));
	      wake_up_interruptible(&frame->wq);
	      // If someone decided to wait for ANY frame - wake him up 
	      if (waitqueue_active(&dev->wq))
	      wake_up_interruptible(&dev->wq);
	    }
	    
	    //read jpeg header data
	    frame = zc030x_next_frame(dev, cData);
            frame->seq = ((int)cData[8]<<8) | (int)cData[9];
	    frame->width = ((int)cData[12]<<8) | (int)cData[13];
	    frame->height = ((int)cData[14]<<8) | (int)cData[15];
            frame->hdrwidth = frame->width;
	    frame->hdrheight = frame->height;
        // FrameWidth = frame->width;
        // FrameHeight = frame->height;
            PDEBUG(5,"***SOI FOUND, seqno:%d, width:%d, height:%d",
			frame->seq, frame->width, frame->height);

       } 

       pData=cData;
       
       memcpy(frame->highwater, pData, datalength);
       frame->highwater += datalength;
       totlen += datalength;
     }	    

     // unLock the buffers
     up(&dev->buf_lock);
     return(totlen);
}

/**
  * zc030x_isoc_handler 
  *
  * isoc callback to move the data received from the iso
  * pipe - it gets linked in isoc_init
  * 
  * 
  * TODO: I think more check needed and some spinlock code
  */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
static void zc030x_isoc_handler (struct urb *urb)
#else
static void zc030x_isoc_handler (struct urb *urb,  struct pt_regs *prgx)
#endif
{
	int i; 
        struct usb_zc030x *dev = (struct usb_zc030x *)urb->context;

	if (!dev)
	{
	   PDEBUG (4, "isoc handler: no context!\n");
	   return;
	}

        if (!dev->streaming) 
	{
           PDEBUG(4, "isoc handler: not streaming, but got interrupt");
	   return;
	}

        /* copy received data to a frame buffer */
        zc030x_move_data(dev, urb);

	dev->cursbuf = (dev->cursbuf + 1) % MAX_ISO_BUFS;

      
        //      IS THIS NEEDED? SOME CODE I FOUND DO THAT
        //
        //      urb->dev = dev->udev;
        //      for (i = 0; i < ISO_FRAMES_PER_DESC; i++) {
        //         urb->iso_frame_desc[i].status = 0;
        //      }
        //      urb->context = dev;
        //      urb->status = 0;
        
	
	i = 0;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0) // wtf does this?
	urb->dev = dev->udev;
	if ((i = usb_submit_urb(urb, GFP_ATOMIC)) != 0) {
	   PDEBUG(4, "Err usb_submit_urb ret %d\n",  i);
	}
#endif
}


/**
  * zc030x_isoc_init() 
  * 
  * start isoc transfer (link handler) 
  * and
  * TODO: start the isoc machine in the camera (this is done implicitly
  * in the init sequences right now)
  *
  */
static int zc030x_isoc_init (struct usb_zc030x *dev)
{
  int i, j;
  int ret;
  struct urb *urb = NULL;
  struct usb_device *udev = dev->udev;
  
  PDEBUG (5, ">> pwc_isoc_init()\n");

  if (!dev){
	  PDEBUG(1,"error no dev!!");
	  return -ENODEV;
  }
  if (!udev){
	  PDEBUG(1,"error no udev!!!");
	  return -ENODEV;
  }
  if (!(dev->sbuf)){
	  PDEBUG(1,"error no dev->sbuf!!!");
	  return -ENODEV;
  }
  for (i = 0; i < MAX_ISO_BUFS; i++){
	  if (!(dev->sbuf[i].urb)){
		  PDEBUG(1,"error no dev->sbuf[%d].urb!!!",i);
		  return -ENODEV;
	  }
  }
  
  /* set streaming variables */
  dev->cursbuf = 0;
  dev->curframe = 0;
  dev->streaming = 0;
  
  /* init URB structure */
  for (i = 0; i < MAX_ISO_BUFS; i++)
    {
     urb = dev->sbuf[i].urb;
     
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) 
      urb->next = dev->sbuf[(i + 1) % MAX_ISO_BUFS].urb;
#endif
      urb->dev = udev;
      urb->pipe = usb_rcvisocpipe (udev, dev->isocend);
      urb->transfer_flags = USB_ISO_ASAP;
      urb->transfer_buffer = dev->sbuf[i].data;
      urb->transfer_buffer_length = ISO_BUFFER_SIZE;
      urb->complete = zc030x_isoc_handler;
      urb->context = dev;
      urb->start_frame = 0;
      urb->number_of_packets = ISO_FRAMES_PER_DESC;
      urb->interval=1;

      for (j = 0; j < ISO_FRAMES_PER_DESC; j++)
	{
	  urb->iso_frame_desc[j].offset = j * ISO_MAX_FRAME_SIZE;
	  urb->iso_frame_desc[j].length = ISO_MAX_FRAME_SIZE;
	}
    }
  
  /* link iso buffers*/
  for (i = 0; i < MAX_ISO_BUFS; i++)
    {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) 
      ret = usb_submit_urb (dev->sbuf[i].urb);
#else
      ret = usb_submit_urb (dev->sbuf[i].urb, GFP_ATOMIC);
#endif      
      if (ret)
      {
	PDEBUG (5, "isoc_init() submit_urb %d failed with error %d\n", i, ret);
      }
      else
      {
	PDEBUG (5, "URB 0x%p submitted.\n", dev->sbuf[i].urb);
      }
    }

  /* All is done... */
  dev->streaming = 1; 
  
  PDEBUG (3, "<< pwc_isoc_init()\n");
  return (0);
}

/**
  * zc030x_isoc_uninit() 
  * 
  * stop isoc transfer (unlink handler) 
  * 
  */
static int zc030x_isoc_uninit (struct usb_zc030x *dev)
{
  struct urb *urb;
  int i;
  PDEBUG(3,">> zc030x_isoc_uninit()");
  
  PDEBUG (3, "unlinking iso buffers\n");
  for (i = MAX_ISO_BUFS -1; i >= 0; i--) {
      urb = dev->sbuf[i].urb;
      if (urb != 0) {
	  //TODO: if streaming!?!? unlink!?
          if (dev->streaming) {
	      PDEBUG (3, "Unlinking URB %p\n", urb);
	      usb_unlink_urb (urb);
	  }
	  PDEBUG (3, "Freeing URB\n");
	  usb_free_urb (urb);
	  dev->sbuf[i].urb = NULL;
	}
    }
  return 0;
}


/***************************************************************************
 *
 * 
 * Device memory allocation & deallocation 
 *
 *
 **************************************************************************/
 

static int zc030x_allocate_buffers (struct usb_zc030x *dev)
{
  int i;
  int s_id;  
  void *kbuf;
  struct urb *urb = NULL;


  PDEBUG (4, ">> zc030x_allocate_buffers(dev = 0x%p)\n", dev);
  down(&dev->buf_lock);
  
  if (dev == NULL)
    return -ENXIO;
  
  s_id = dev->sensor_id; //sensor id
  
  /* Allocate Isochronuous pipe buffers */
  for (i = 0; i < MAX_ISO_BUFS; i++) {
    if (dev->sbuf[i].data == NULL){
          /* allocate data buffer */	
	  kbuf = kmalloc (ISO_BUFFER_SIZE, GFP_KERNEL);
	  if (kbuf == NULL){
	      PDEBUG (1, "Failed to allocate data for iso buffer %d.", i);
	      goto error; /* NO MEMORY */
	  }
	  PDEBUG (4, "Allocated iso buffer at %p.", kbuf);
	  memset (kbuf, 0, ISO_BUFFER_SIZE);
	  dev->sbuf[i].data = kbuf;
	  /* allocate urb */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) 
          urb = usb_alloc_urb(ISO_FRAMES_PER_DESC);
#else
          urb = usb_alloc_urb(ISO_FRAMES_PER_DESC, GFP_KERNEL);	  
#endif
          if (urb == NULL){
	      PDEBUG (1, "Failed to allocate urb for iso buffer %d", i);
	      goto error; /* NU MEMURY */
          }
	  PDEBUG (4, "Allocated URB at %p.", urb);
	  dev->sbuf[i].urb = urb;
	}
  }
 
  /* allocate video buffer space */
  dev->fbuf = rvmalloc(ZC030X_NUMFRAMES * MAX_VIDEODATA_SIZE);
  if (!dev->fbuf) {
      PDEBUG (1, "Failed to allocate iso buffer %d.\n", i);
      goto error; //NO MEMORY
  }
  
  dev->mmbuf = rvmalloc(MAX_VIDEODATA_SIZE);
  if (!dev->mmbuf) {
      PDEBUG (1, "Failed to allocate mmap buffer %d.\n", i);
      goto error; //NO MEMORY
  }

  /* Allocate temporary buffer for jpeg decoder */
  dev->tmpBuffer = rvmalloc(MAX_VIDEODATA_SIZE);
  if(!dev->tmpBuffer)
    goto error; //NO MEMORY

  /* init frame structure */
  for (i = 0; i < ZC030X_NUMFRAMES; i++) 
  {
    dev->frame[i].tmpbuffer = dev->tmpBuffer;
    dev->frame[i].grabstate = FRAME_UNUSED;
    dev->frame[i].scanstate = STATE_SCANNING;
    dev->frame[i].data = dev->fbuf + i * MAX_VIDEODATA_SIZE; 
    dev->frame[i].highwater=dev->frame[i].data;
    PDEBUG(4, "frame[%d] @ %p", i, dev->frame[i].data);
  }

  up(&dev->buf_lock);
  PDEBUG (4, "<< zc030x_allocate_buffers.\n");
  return 0;

error:
    
  for (i = 0; i < MAX_ISO_BUFS; i++) {
    if (dev->sbuf[i].data) {
      kfree(dev->sbuf[i].data);
      dev->sbuf[i].data = NULL;
    }
    if (dev->sbuf[i].urb) {
      usb_free_urb(dev->sbuf[i].urb);
      dev->sbuf[i].urb = NULL;
    }
  }
  if (dev->fbuf) {
    rvfree(dev->fbuf, ZC030X_NUMFRAMES * MAX_VIDEODATA_SIZE);
    dev->fbuf = NULL;
  }
  if (dev->tmpBuffer) {
    rvfree(dev->tmpBuffer,MAX_VIDEODATA_SIZE);
    dev->tmpBuffer = NULL;
  }
    
  dev->buf_state = BUF_NOT_ALLOCATED;
  up(&dev->buf_lock);
  PDEBUG(1, "errored");
  return -ENOMEM;
}


static int zc030x_unallocate_buffers(struct usb_zc030x *dev)
{
  int i;
  int s_id;
 
  PDEBUG(3,"zc030x_unallocate_buffers()");
  
  down(&dev->buf_lock);
  
  if (dev == NULL)
     return -ENXIO;
  s_id = dev->sensor_id;

  for (i = 0; i < MAX_ISO_BUFS; i++) {
    if (dev->sbuf[i].data != NULL) {
      kfree(dev->sbuf[i].data);
      dev->sbuf[i].data = NULL;
    }
    if (dev->sbuf[i].urb != NULL) {
      usb_free_urb(dev->sbuf[i].urb);
      dev->sbuf[i].urb = NULL;
    }
  }
    
  if (dev->fbuf) {
    rvfree(dev->fbuf, ZC030X_NUMFRAMES * MAX_VIDEODATA_SIZE);
    dev->fbuf = NULL;
  }
  if (dev->tmpBuffer) {
    rvfree(dev->tmpBuffer,MAX_VIDEODATA_SIZE);
    dev->tmpBuffer = NULL;
  }

  dev->buf_state = BUF_NOT_ALLOCATED;
  
  up(&dev->buf_lock);
  return 0;
}


/***************************************************************************
 *
 * 
 * Camera detection & initialization functions 
 * (TODO: needs alot of reorganization)
 *
 *
 **************************************************************************/

/* FIXME to be implemented */
static int zc030x_reg_read(struct usb_device *udev, __u16 reg, __u16 index, __u16 length)
{
    int rc = 0;
    return rc;
}


/* FIXME to be implemented */
/* write a register in the zc030x chip */
static int zc030x_reg_write(struct usb_device *udev, __u16 reg, __u16 index, __u16 value)
{
    int rc;

    rc = usb_control_msg(udev, 
		         usb_sndctrlpipe(udev, 0), 
			 reg,
			 USB_TYPE_VENDOR | USB_RECIP_DEVICE,
			 value, 
			 index, 
			 NULL, 
			 0, 
			 HZ);

    PDEBUG(5, "reg write: 0x%02X,0x%02X:0x%02X, 0x%x", reg, index, value, rc);

   if (rc < 0)
         err("reg write: error %d", rc);

   return rc;
}

/* Try to read an I2C register */
static int zc030x_i2c_reg_read(struct usb_device *udev, __u16 addr, __u16 size)
{
    int ret = 0;
    unsigned char buf[4] = { 0,0,0,0 };
    
    PDEBUG(2, "<<< I2CREGREAD : udev = %p, addr %04X(%d)", udev, addr, size);
    
    // Select the register now...
    ret = usb_control_msg (udev, usb_sndctrlpipe (udev, 0),
             0x00A0, 0x40, addr, 0x0092, buf, 0x0000, 5000);
    ret = usb_control_msg (udev, usb_sndctrlpipe (udev, 0),
             0x00A0, 0x40, 0x0002, 0x0090, buf, 0x0000, 5000);
    ret = usb_control_msg (udev, usb_rcvctrlpipe (udev, 0), 
             0x00A1, 0xc0, 0x0001, 0x0091, buf, 1, 5000);
    ret = usb_control_msg (udev, usb_rcvctrlpipe (udev, 0), 
             0x00A1, 0xc0, 0x0001, 0x0095, buf, size, 5000);
    PDEBUG(2, ">>> I2CREGREAD : ret %d, value : %04X", ret, (buf[0]<<8)|(buf[1]));
    return (*(int*)buf);
}

/* Try to write an I2C register */
static int zc030x_i2c_reg_write(struct usb_device *udev, __u16 addr, __u16 value)
{
    int ret = 0;
    unsigned char buf[20];
    
    PDEBUG(2, "<<< I2CREGWRITE : udev = %p, addr %04X, value %04X", udev, addr, value);
    
    // Select the register now...
    ret = usb_control_msg (udev, usb_sndctrlpipe (udev, 0),
             0x00A0, 0x40, addr, 0x0092, buf, 0x0000, 5000);
    ret = usb_control_msg (udev, usb_sndctrlpipe (udev, 0),
             0x00A0, 0x40, value, 0x0093, buf, 0x0000, 5000);
    ret = usb_control_msg (udev, usb_sndctrlpipe (udev, 0),
             0x00A0, 0x40, 0x0000, 0x0094, buf, 0x0000, 5000);
    ret = usb_control_msg (udev, usb_sndctrlpipe (udev, 0),
             0x00A0, 0x40, 0x0001, 0x0090, buf, 0x0000, 5000);
    ret = usb_control_msg (udev, usb_rcvctrlpipe (udev, 0), 
             0x00A1, 0xc0, 0x0001, 0x0091, buf, 0x0001, 5000);
    PDEBUG(2, ">>> I2CREGWRITE : ret %d, value : %02X", ret, buf[0]);
    return 0;
}

/* Decode sequences from the windows .inf file */
void
send_seq (struct usb_device *udev, __u8 * data, int size)
{
  unsigned char buf[1];
  int i;
  int index, value;
  int ret;

  for (i = 0; i < size; i += 4)
    {
      switch (data[i + 3])
	{
	default:
	case 0xcc:
	  index = data[i] * 0x100 + data[i + 1];
	  value = data[i + 2];
	  ret =
	    usb_control_msg (udev, usb_sndctrlpipe (udev, 0),
			     0x00A0, 0x40, value, index, buf, 0x0000, HZ);
	  break;
	case 0xaa:
	  buf[0] = 0x00;
	  ret =
	    usb_control_msg (udev, usb_sndctrlpipe (udev, 0),
			     0x00A0, 0x40, data[i + 1], 0x0092, buf, 0x0000,
			     HZ);
	  ret =
	    usb_control_msg (udev, usb_sndctrlpipe (udev, 0), 0x00A0, 0x40,
			     data[i + 2], 0x0093, buf, 0x0000, HZ);
	  ret =
	    usb_control_msg (udev, usb_sndctrlpipe (udev, 0), 0x00A0, 0x40,
			     data[i], 0x0094, buf, 0x0000, HZ);
	  ret =
	    usb_control_msg (udev, usb_sndctrlpipe (udev, 0), 0x00A0, 0x40,
			     0x0001, 0x0090, buf, 0x0000, HZ);
	  ret =
	    usb_control_msg (udev, usb_rcvctrlpipe (udev, 0), 0x00A1, 0xC0,
			     0x0001, 0x0091, buf, 0x0001, HZ);
	  PDEBUG
	    (5,"submitting(recv): 0xc0 0xa1 0x0001 0x0091 1 return:%d data:%2x",
	     ret, buf[0]);
	  break;
	case 0xdd:
	  /* Sequence present in w300a seq but meaning unknown */
	  break;
	}
    }
}

/*
 * send/receive urb-sequence-arrays (in header files) to/from device  
 */

static void zc030x_playbackControlSeq (struct usb_device *udev, struct usb_zc030x *dev, __u16 seq[][5], int length)
{
  unsigned char buf[20];
  int ret;
  int i;

  //urb vars
  unsigned char urb_request;
  unsigned char urb_requesttype;
  unsigned int urb_value;
  unsigned int urb_index;
  unsigned int urb_size;

  for (i = 0; i < length; i++)
  {
     urb_requesttype = (unsigned char) (seq[i][0]);
     urb_request = (unsigned char) (seq[i][1]);
     urb_value = (unsigned int) (seq[i][2]);
     urb_index = (unsigned int) (seq[i][3]);
     urb_size = (unsigned int) (seq[i][4]);

     if ((urb_requesttype & 0x80) > 0)
     // device -> host
     {
       // recv control message         
       ret = usb_control_msg (udev, usb_rcvctrlpipe (udev, 0), urb_request,
				 urb_requesttype, urb_value, urb_index, 
				 buf, urb_size, 5000);

       PDEBUG(5, "submitting(recv): %2x %2x %4x %4x %4x urb:%d return:%d data:%2x",
	     urb_requesttype, urb_request, urb_value, urb_index, urb_size, i,
	     ret, buf[0]);
     }
     else
     // host -> device
     {
       // send control message         
       ret = usb_control_msg (udev, usb_sndctrlpipe (udev, 0), urb_request,
				 urb_requesttype, urb_value, urb_index,
				 buf, urb_size, 5000);

       PDEBUG (5, "submitting(send): %2x %2x %4x %4x %4x urb:%d return:%d",
	      urb_requesttype, urb_request, urb_value, urb_index, urb_size,
		i, ret);
     }
  }

}

static void reenactSequence(struct usb_device *udev, __u16 sequence[][5], int length)
{
  int ret, i;
  unsigned char urb_request, urb_requesttype;
  unsigned int urb_value, urb_index, urb_size;
  unsigned char buf[20];

  for (i = 0; i < length; i++)
    {

      urb_requesttype = (unsigned char) (sequence[i][0]);
      urb_request = (unsigned char) (sequence[i][1]);
      urb_value = (unsigned int) (sequence[i][2]);
      urb_index = (unsigned int) (sequence[i][3]);
      urb_size = (unsigned int) (sequence[i][4]);


      if ((urb_requesttype & 0x80) > 0)
	// device -> host
	{
	  // recv control message         
	  ret = usb_control_msg (udev,
				 usb_rcvctrlpipe (udev, 0),
				 urb_request,
				 urb_requesttype,
				 urb_value, urb_index, buf, urb_size, HZ);

	  PDEBUG
	    (5,"submitting(recv): %2x %2x %4x %4x %4x urb:%d return:%d data:%2x",
	     urb_requesttype, urb_request, urb_value, urb_index, urb_size, i,
	     ret, buf[0]);
	}
      else
	// host -> device
	{
	  // send control message         
	  ret = usb_control_msg (udev,
				 usb_sndctrlpipe (udev, 0),
				 urb_request,
				 urb_requesttype,
				 urb_value, urb_index, buf, urb_size, HZ);

	  PDEBUG (5,"submitting(send): %2x %2x %4x %4x %4x urb:%d return:%d",
		  urb_requesttype, urb_request, urb_value, urb_index, urb_size,
		  i, ret);
	}

    }

}



static int
webcamnxpro_init (struct usb_device *udev, struct usb_zc030x *dev)
{
  reenactSequence(udev, nxpro_plugin, 19);
  reenactSequence(udev, nxpro_alt_init, 108);
  reenactSequence(udev, nxpro_gammat, 32);
  reenactSequence(udev, nxpro_matrixt, 9);
  reenactSequence(udev, nxpro_unknown1, 2);
  reenactSequence(udev, nxpro_60hz, 50);
  reenactSequence(udev, nxpro_unknown2, 11);
  usb_set_interface (udev, 0, 6);
  return 0;
}

static int webcamnotebook_init (struct usb_device *udev, struct usb_zc030x *dev)
{

  usb_set_interface (udev, 0, 0);
  reenactSequence(udev, webcamnotebook_init1,24);
  reenactSequence(udev, webcamnotebook_InitialScale,33);
  reenactSequence(udev, webcamnotebook_unknown1,8);
  reenactSequence(udev, webcamnotebook_gamma,32);
  reenactSequence(udev, webcamnotebook_matrix,9);
  reenactSequence(udev, webcamnotebook_unknown2, 4);
  reenactSequence(udev, webcamnotebook_50HzScale, 29);
  reenactSequence(udev, webcamnotebook_unknonw3, 4);

  
  usb_set_interface (udev, 0, 6);
  return 0;


}
static int
webcamnx_init (struct usb_device *udev, struct usb_zc030x *dev)
{
  unsigned char buf[20];
  int ret;
  int i;

  //__u8 d[] = { 0x41, 0x00, 0x00, 0x00, 0x20, 0x90, 0x8F, 0xc7 };
  //urb vars
  unsigned char urb_request;
  unsigned char urb_requesttype;
  unsigned int urb_value;
  unsigned int urb_index;
  unsigned int urb_size;


  //
  // send plugin sequence
  //

  usb_set_interface (udev, 0, 0);

  for (i = 0; i < 116; i++)
    {

      urb_requesttype = (unsigned char) (webcamnx_init1[i][0]);
      urb_request = (unsigned char) (webcamnx_init1[i][1]);
      urb_value = (unsigned int) (webcamnx_init1[i][2]);
      urb_index = (unsigned int) (webcamnx_init1[i][3]);
      urb_size = (unsigned int) (webcamnx_init1[i][4]);


      if ((urb_requesttype & 0x80) > 0)
	// device -> host
	{
	  // recv control message         
	  ret = usb_control_msg (udev,
				 usb_rcvctrlpipe (udev, 0),
				 urb_request,
				 urb_requesttype,
				 urb_value, urb_index, buf, urb_size, HZ);

	  PDEBUG
	    (5,"submitting(recv): %2x %2x %4x %4x %4x urb:%d return:%d data:%2x",
	     urb_requesttype, urb_request, urb_value, urb_index, urb_size, i,
	     ret, buf[0]);
	}
      else
	// host -> device
	{
	  // send control message         
	  ret = usb_control_msg (udev,
				 usb_sndctrlpipe (udev, 0),
				 urb_request,
				 urb_requesttype,
				 urb_value, urb_index, buf, urb_size, HZ);

	  PDEBUG (5,"submitting(send): %2x %2x %4x %4x %4x urb:%d return:%d",
		urb_requesttype, urb_request, urb_value, urb_index, urb_size,
		i, ret);
	}

    }

  send_seq (udev, pas106b_Initial, sizeof (pas106b_Initial));

  for (i = 0; i < 9; i++)
    {

      urb_requesttype = (unsigned char) (webcamnx_unknown1[i][0]);
      urb_request = (unsigned char) (webcamnx_unknown1[i][1]);
      urb_value = (unsigned int) (webcamnx_unknown1[i][2]);
      urb_index = (unsigned int) (webcamnx_unknown1[i][3]);
      urb_size = (unsigned int) (webcamnx_unknown1[i][4]);


      if ((urb_requesttype & 0x80) > 0)
	// device -> host
	{
	  // recv control message         
	  ret = usb_control_msg (udev,
				 usb_rcvctrlpipe (udev, 0),
				 urb_request,
				 urb_requesttype,
				 urb_value, urb_index, buf, urb_size, HZ);

	  PDEBUG 
	    (5,"submitting(recv): %2x %2x %4x %4x %4x urb:%d return:%d data:%2x",
	     urb_requesttype, urb_request, urb_value, urb_index, urb_size, i,
	     ret, buf[0]);
	}
      else
	// host -> device
	{
	  // send control message         
	  ret = usb_control_msg (udev,
				 usb_sndctrlpipe (udev, 0),
				 urb_request,
				 urb_requesttype,
				 urb_value, urb_index, buf, urb_size, HZ);

	  PDEBUG (5,"submitting(send): %2x %2x %4x %4x %4x urb:%d return:%d",
		urb_requesttype, urb_request, urb_value, urb_index, urb_size,
		i, ret);
	}

    }
  send_seq (udev, pas106b_GammaT, sizeof (pas106b_GammaT));
  send_seq (udev, pas106b_MatrixT, sizeof (pas106b_MatrixT));
  for (i = 0; i < 2; i++)
    {

      urb_requesttype = (unsigned char) (webcamnx_unknown2[i][0]);
      urb_request = (unsigned char) (webcamnx_unknown2[i][1]);
      urb_value = (unsigned int) (webcamnx_unknown2[i][2]);
      urb_index = (unsigned int) (webcamnx_unknown2[i][3]);
      urb_size = (unsigned int) (webcamnx_unknown2[i][4]);


      if ((urb_requesttype & 0x80) > 0)
	// device -> host
	{
	  // recv control message         
	  ret = usb_control_msg (udev,
				 usb_rcvctrlpipe (udev, 0),
				 urb_request,
				 urb_requesttype,
				 urb_value, urb_index, buf, urb_size, HZ);

	 PDEBUG 
	    (5, "submitting(recv): %2x %2x %4x %4x %4x urb:%d return:%d data:%2x",
	     urb_requesttype, urb_request, urb_value, urb_index, urb_size, i,
	     ret, buf[0]);
	}
      else
	// host -> device
	{
	  // send control message         
	  ret = usb_control_msg (udev,
				 usb_sndctrlpipe (udev, 0),
				 urb_request,
				 urb_requesttype,
				 urb_value, urb_index, buf, urb_size, HZ);

	  PDEBUG (5, "submitting(send): %2x %2x %4x %4x %4x urb:%d return:%d",
		urb_requesttype, urb_request, urb_value, urb_index, urb_size,
		i, ret);
	}

    }
  send_seq (udev, pas106b_60HZ, sizeof (pas106b_60HZ));
  for (i = 0; i < 7; i++)
    {

      urb_requesttype = (unsigned char) (webcamnx_unknown3[i][0]);
      urb_request = (unsigned char) (webcamnx_unknown3[i][1]);
      urb_value = (unsigned int) (webcamnx_unknown3[i][2]);
      urb_index = (unsigned int) (webcamnx_unknown3[i][3]);
      urb_size = (unsigned int) (webcamnx_unknown3[i][4]);


      if ((urb_requesttype & 0x80) > 0)
	// device -> host
	{
	  // recv control message         
	  ret = usb_control_msg (udev,
				 usb_rcvctrlpipe (udev, 0),
				 urb_request,
				 urb_requesttype,
				 urb_value, urb_index, buf, urb_size, HZ);

	  PDEBUG 
	    (5,"submitting(recv): %2x %2x %4x %4x %4x urb:%d return:%d data:%2x",
	     urb_requesttype, urb_request, urb_value, urb_index, urb_size, i,
	     ret, buf[0]);
	}
      else
	// host -> device
	{
	  // send control message         
	  ret = usb_control_msg (udev,
				 usb_sndctrlpipe (udev, 0),
				 urb_request,
				 urb_requesttype,
				 urb_value, urb_index, buf, urb_size, HZ);

	  PDEBUG (5, "submitting(send): %2x %2x %4x %4x %4x urb:%d return:%d",
		urb_requesttype, urb_request, urb_value, urb_index, urb_size,
		i, ret);
	}

    }
  usb_set_interface (udev, 0, 6);
  return ret;
}

static int generic_init (struct usb_device *udev, struct usb_zc030x *dev, __u8 * InitialArray, __u8 * InitialScaleArray, __u8 * GammaTArray, __u8 * MatrixTArray, __u8 * FreqArray)
{
  //unsigned char buf[20];
  unsigned char buf[20];
  int i;

  //__u8 d[] = { 0x41, 0x00, 0x00, 0x00, 0x20, 0x90, 0x8F, 0xc7 };
  //urb vars
  unsigned char urb_request;
  unsigned char urb_requesttype;
  unsigned int urb_value;
  unsigned int urb_index;
  unsigned int urb_size;

  int ret = 0;
  //int i;
  //__u8 d[] = { 0x00, 0x00, 0x00, 0x00, 0x40, 0xd6, 0x97, 0xf8 };

  //
  // send plugin sequence
  //

  usb_set_interface (udev, 0, 0);

  // Plugin sequence (I've no access to this)
  zc030x_playbackControlSeq (udev, dev, wcam300a_init1, 144);
	  
  usb_set_interface (udev, 0, 0);

  // Initial sequence
  send_seq (udev, InitialArray, sizeof (InitialArray));

  //set default alternate interface for wcam300a (differs for cam models!!)
  usb_set_interface (udev, 0, 6);

  //this is the "initialScale" sequence
  // Initial sequence
  send_seq (udev, InitialScaleArray, sizeof (InitialScaleArray));
  
  //this is the "50HzScale" sequence 
  send_seq (udev, FreqArray, sizeof (FreqArray));
  
  return ret;


  //
  // send plugin sequence
  //

  usb_set_interface (udev, 0, 0);



  for (i = 0; i < 9; i++)
    {

      urb_requesttype = (unsigned char) (webcamnx_unknown1[i][0]);
      urb_request = (unsigned char) (webcamnx_unknown1[i][1]);
      urb_value = (unsigned int) (webcamnx_unknown1[i][2]);
      urb_index = (unsigned int) (webcamnx_unknown1[i][3]);
      urb_size = (unsigned int) (webcamnx_unknown1[i][4]);


      if ((urb_requesttype & 0x80) > 0)
	// device -> host
	{
	  // recv control message         
	  ret = usb_control_msg (udev,
				 usb_rcvctrlpipe (udev, 0),
				 urb_request,
				 urb_requesttype,
				 urb_value, urb_index, buf, urb_size, HZ);

	  PDEBUG 
	    (5,"submitting(recv): %2x %2x %4x %4x %4x urb:%d return:%d data:%2x",
	     urb_requesttype, urb_request, urb_value, urb_index, urb_size, i,
	     ret, buf[0]);
	}
      else
	// host -> device
	{
	  // send control message         
	  ret = usb_control_msg (udev,
				 usb_sndctrlpipe (udev, 0),
				 urb_request,
				 urb_requesttype,
				 urb_value, urb_index, buf, urb_size, HZ);

	  PDEBUG (5,"submitting(send): %2x %2x %4x %4x %4x urb:%d return:%d",
		urb_requesttype, urb_request, urb_value, urb_index, urb_size,
		i, ret);
	}

    }
  send_seq (udev, GammaTArray, sizeof (GammaTArray));
  send_seq (udev, MatrixTArray, sizeof (MatrixTArray));
  for (i = 0; i < 2; i++)
    {

      urb_requesttype = (unsigned char) (webcamnx_unknown2[i][0]);
      urb_request = (unsigned char) (webcamnx_unknown2[i][1]);
      urb_value = (unsigned int) (webcamnx_unknown2[i][2]);
      urb_index = (unsigned int) (webcamnx_unknown2[i][3]);
      urb_size = (unsigned int) (webcamnx_unknown2[i][4]);


      if ((urb_requesttype & 0x80) > 0)
	// device -> host
	{
	  // recv control message         
	  ret = usb_control_msg (udev,
				 usb_rcvctrlpipe (udev, 0),
				 urb_request,
				 urb_requesttype,
				 urb_value, urb_index, buf, urb_size, HZ);

	 PDEBUG 
	    (5, "submitting(recv): %2x %2x %4x %4x %4x urb:%d return:%d data:%2x",
	     urb_requesttype, urb_request, urb_value, urb_index, urb_size, i,
	     ret, buf[0]);
	}
      else
	// host -> device
	{
	  // send control message         
	  ret = usb_control_msg (udev,
				 usb_sndctrlpipe (udev, 0),
				 urb_request,
				 urb_requesttype,
				 urb_value, urb_index, buf, urb_size, HZ);

	  PDEBUG (5, "submitting(send): %2x %2x %4x %4x %4x urb:%d return:%d",
		urb_requesttype, urb_request, urb_value, urb_index, urb_size,
		i, ret);
	}

    }
  send_seq (udev, FreqArray, sizeof (FreqArray));
  for (i = 0; i < 7; i++)
    {

      urb_requesttype = (unsigned char) (webcamnx_unknown3[i][0]);
      urb_request = (unsigned char) (webcamnx_unknown3[i][1]);
      urb_value = (unsigned int) (webcamnx_unknown3[i][2]);
      urb_index = (unsigned int) (webcamnx_unknown3[i][3]);
      urb_size = (unsigned int) (webcamnx_unknown3[i][4]);


      if ((urb_requesttype & 0x80) > 0)
	// device -> host
	{
	  // recv control message         
	  ret = usb_control_msg (udev,
				 usb_rcvctrlpipe (udev, 0),
				 urb_request,
				 urb_requesttype,
				 urb_value, urb_index, buf, urb_size, HZ);

	  PDEBUG 
	    (5,"submitting(recv): %2x %2x %4x %4x %4x urb:%d return:%d data:%2x",
	     urb_requesttype, urb_request, urb_value, urb_index, urb_size, i,
	     ret, buf[0]);
	}
      else
	// host -> device
	{
	  // send control message         
	  ret = usb_control_msg (udev,
				 usb_sndctrlpipe (udev, 0),
				 urb_request,
				 urb_requesttype,
				 urb_value, urb_index, buf, urb_size, HZ);

	  PDEBUG (5, "submitting(send): %2x %2x %4x %4x %4x urb:%d return:%d",
		urb_requesttype, urb_request, urb_value, urb_index, urb_size,
		i, ret);
	}

    }
  usb_set_interface (udev, 0, 6);
  return ret;
}


static int
zc030x_init (struct usb_device *udev, struct usb_zc030x *dev)
{
  //unsigned char buf[20];
  int ret = 0;
  //int i;
  //__u8 d[] = { 0x00, 0x00, 0x00, 0x00, 0x40, 0xd6, 0x97, 0xf8 };

  //
  // send plugin sequence
  //

  usb_set_interface (udev, 0, 0);

  zc030x_playbackControlSeq (udev, dev, wcam300a_init1, 144);
	  
  usb_set_interface (udev, 0, 0);

  zc030x_playbackControlSeq (udev, dev, zc030x_seq0, 35);
  zc030x_playbackControlSeq (udev, dev, zc030x_seq1, 4);
  zc030x_playbackControlSeq (udev, dev, zc030x_seq2, 3);
  zc030x_playbackControlSeq (udev, dev, zc030x_seq3, 7);
  zc030x_playbackControlSeq (udev, dev, zc030x_seq4, 16);
  zc030x_playbackControlSeq (udev, dev, zc030x_seq5, 16);
  zc030x_playbackControlSeq (udev, dev, zc030x_seq6, 9);
  zc030x_playbackControlSeq (udev, dev, zc030x_seq7, 2);
  zc030x_playbackControlSeq (udev, dev, zc030x_seq8, 30);
  zc030x_playbackControlSeq (udev, dev, zc030x_seq9, 3);
  zc030x_playbackControlSeq (udev, dev, zc030x_seq10, 3);

  //set default alternate interface for wcam300a (differs for cam models!!)
  usb_set_interface (udev, 0, 6);

  //this is the "initialScale" sequence
  zc030x_playbackControlSeq (udev, dev, zc030x_seq0, 35);
  
  zc030x_playbackControlSeq (udev, dev, zc030x_seq1, 4);
  zc030x_playbackControlSeq (udev, dev, zc030x_seq2, 3);
  zc030x_playbackControlSeq (udev, dev, zc030x_seq3, 7);
  zc030x_playbackControlSeq (udev, dev, zc030x_seq4, 16);
  zc030x_playbackControlSeq (udev, dev, zc030x_seq5, 16);
  zc030x_playbackControlSeq (udev, dev, zc030x_seq6, 9);
  zc030x_playbackControlSeq (udev, dev, zc030x_seq7, 2);
  
  //this is the "50HzScale" sequence 
  zc030x_playbackControlSeq (udev, dev, zc030x_seq8, 30);
  zc030x_playbackControlSeq (udev, dev, zc030x_seq9, 3);
  zc030x_playbackControlSeq (udev, dev, zc030x_seq10, 3);
  
  return ret;
}

static int
benqcam_init (struct usb_device *udev, struct usb_zc030x *dev){
//	return zc030x_init(udev, dev);
	zc030x_playbackControlSeq (udev, dev, benqcam_plugin, 184);
	PDEBUG(1, "benqcam init if 6");	
	usb_set_interface (udev, 0, 6);
	zc030x_playbackControlSeq (udev, dev, benqcam_capture, 325);
	
}

static int
zstar_init (struct usb_device *udev, struct usb_zc030x *dev){
//	return zc030x_init(udev, dev);
	zc030x_playbackControlSeq (udev, dev, zstar_plugin, 104);
	PDEBUG(1, "ZStar init if 6");	
	usb_set_interface (udev, 0, 7);
	zc030x_playbackControlSeq (udev, dev, zstar_capture320, 768);
	
}

static int
webcammobile_init (struct usb_device *udev, struct usb_zc030x *dev)
{
  int ret = 0;

  usb_set_interface (udev, 0, 0);

  // send plugin sequence
  zc030x_playbackControlSeq (udev, dev, webcammobile_plugin, 100);
  	
  zc030x_playbackControlSeq (udev, dev, webcammobile_seq0, 434);
  zc030x_playbackControlSeq (udev, dev, webcammobile_seq1, 4);
  zc030x_playbackControlSeq (udev, dev, webcammobile_seq2, 3);
  zc030x_playbackControlSeq (udev, dev, webcammobile_seq3, 7);
  zc030x_playbackControlSeq (udev, dev, webcammobile_seq4, 32);
  zc030x_playbackControlSeq (udev, dev, webcammobile_seq5, 9 );
  zc030x_playbackControlSeq (udev, dev, webcammobile_seq6, 2);
  zc030x_playbackControlSeq (udev, dev, webcammobile_seq7, 75);
  zc030x_playbackControlSeq (udev, dev, webcammobile_seq8, 3);
  zc030x_playbackControlSeq (udev, dev, webcammobile_seq9, 3);

  usb_set_interface (udev, 0, 6);
	
  return ret;
}

/**
 *	zc030x_detectCamera
 *
 *	Called by zc030x_probe when a new device is connected
 *
 *      TODO: this function is pretty simple right now,
 *            maybe there is a possibility to detect the
 *            bridge/sensor - this should go here
 */

static int zc030x_detect_camera(struct usb_zc030x *dev)
{
  
  int type_id;
  int sensor_id;
  struct usb_device *udev = dev->udev;

  int vend_id = udev->descriptor.idVendor;
  int prod_id = udev->descriptor.idProduct;
  int release = udev->descriptor.bcdDevice;
  /* get release number */

  /* some info about the device */
  PDEBUG (1, "vend_id : %04x \n", vend_id);
  PDEBUG (1, "prod_id : %04x \n", prod_id);
  PDEBUG (1, "Release : %04x \n", release);

  /*   See if the device offered us matches what we can accept */
 
  type_id = UNKNOWN;
  sensor_id = UNKNOWN;
  switch (vend_id)
    {
    case USB_W300A_VENDOR_ID:
      if (prod_id == USB_W300A_PRODUCT_ID)
      {
	type_id = W300A;
	sensor_id = TAS5130C; 
      }
      break;
    case USB_WEBCAMNXPRO_VENDOR_ID:
      if (prod_id == USB_WEBCAMNX_PRODUCT_ID)
      {
	type_id = WEBCAMNX;
	sensor_id = PAS106B;
      }
      if (prod_id == USB_WEBCAMNXPRO_PRODUCT_ID)
      {
	type_id = WEBCAMNXPRO;
        sensor_id = HV7131B; 
      }
      if (prod_id == USB_WEBCAMMOBILE_PRODUCT_ID)
      {
	type_id = WEBCAMMOBILE;
	//sensor_id = ICM105A;
      }
      if (prod_id == USB_WEBCAMNOTEBOOK_PRODUCT_ID)
      {
	type_id = WEBCAMNOTEBOOK;
	sensor_id = TAS5130C;
      }
      break;
    case USB_GENIUSV2_VENDOR_ID:
      if (prod_id == USB_GENIUSV2_PRODUCT_ID)
      {
	type_id = GENIUSV2;
        sensor_id = TAS5130C;
      }
      break;
    case USB_BENQCAM_VENDOR_ID:
      if (prod_id == USB_BENQCAM_PRODUCT_ID)
      {
	type_id = BENQCAM;
	sensor_id = HV7131B;
      }
      if (prod_id == USB_ZSTARCORP_PRODUCT_ID)
      {
	type_id = ZSTARCORP;
	sensor_id = TAS5130C;
      }
      break;
    default:
      break;
    }
  if (type_id == UNKNOWN)
  {
     PDEBUG (1, "Unsupported webcam. Exiting\n");
     return -1;
  }

  PDEBUG (1, "Detected cam %s\n", clist[type_id].name);

  // set the parameters we've found out about the device
  // (TODO: maybe also sensortype in future)
  dev->vend_id = vend_id;
  dev->prod_id = prod_id;
  dev->type_id = type_id;
  dev->sensor_id = sensor_id;
  //hardwired alt_setting bad?
  dev->alt = default_alt_setting[type_id];
  return 0;
}


/**
 *	zc030x_configure_camera
 *
 *	Called by zc030x_probe when a new device is connected
 *
 *      TODO: this function is pretty simple right now,
 *            configuration/sensor initialization
 *            should be done here according to camera type/sensor type...
 *            
 *            Right now the old init is done....
 *      TODO: return apropriate error codes on fail. 
 */

static int zc030x_configure_camera(struct usb_zc030x *dev)
{
  struct usb_device *udev = dev->udev;
  
  PDEBUG (2, "initializing camera... ");
  switch (dev->type_id)
    {
    case W300A:
	memcpy(&dev->params,&TAS5130C_defaultSettings,sizeof(struct cam_params));
	zc030x_allocate_buffers (dev);
	zc030x_init (udev, dev);
	break;
    case WEBCAMNX:
	memcpy(&dev->params,&PAS106B_defaultSettings,sizeof(struct cam_params));
	zc030x_allocate_buffers (dev);
	webcamnx_init (udev, dev);
	break;
    case WEBCAMNXPRO:
	memcpy(&dev->params,&HV7131B_defaultSettings,sizeof(struct cam_params));
	zc030x_allocate_buffers (dev);
	webcamnxpro_init (udev, dev);
        break;
    case WEBCAMMOBILE:
	memcpy(&dev->params,&ICM105A_defaultSettings,sizeof(struct cam_params));
	zc030x_allocate_buffers (dev);
      	webcammobile_init (udev, dev);
	break;
    case WEBCAMNOTEBOOK:
	memcpy(&dev->params,&TAS5130C_defaultSettings,sizeof(struct cam_params));
	zc030x_allocate_buffers (dev);
	webcamnotebook_init (udev, dev);
	break;
    case BENQCAM:
	memcpy(&dev->params,&HV7131B_defaultSettings,sizeof(struct cam_params));
	zc030x_allocate_buffers (dev);
	benqcam_init (udev, dev);
	break;
    case GENIUSV2:
      zc030x_allocate_buffers (dev);
      zc030x_init (udev, dev);
	break;
    case ZSTARCORP:
//	memcpy(&dev->params,&TAS5130C_defaultSettings,sizeof(struct cam_params));
	zc030x_allocate_buffers (dev);
//	generic_init (udev, dev, TAS5130C_ZStar_Initial, TAS5130C_ZStar_InitialScale,  TAS5130C_GammaT,  TAS5130C_MatrixT,  TAS5130C_50HZ);
	zstar_init (udev, dev);
    }
  
  return 0;
} 



/*************************************************************************
 *
 * 
 *  Driver File operations (FOPS)                                        
 *  
 *  
 ************************************************************************/

/**
  * 
  *	zc030x_open
  *
  * TODO: Review if isoc init can be called here, and if stream is stopped correctly on close
  *       memory allocation/deallocation can also be done on open/close (thanks Andre)  - but don't know yet
  *       how this will work together with v4l_mmap (which is needed for xawtv), is there an open before xawtv reads?
  *       If yes alloc/dealloc & isoc init can be all done in open() and undone in close().
  * 
  */
static int zc030x_open (struct inode *inode, struct file *file)
{
  struct video_device *vdev = video_devdata(file);
  struct usb_zc030x *dev;
  int retval = 0;

  PDEBUG(3,">> zc030x_open()");
  
  if (vdev == NULL)
  {
    PDEBUG (2, "vdev is NULL, returning");
    return 0;
  }

  dev = video_get_drvdata(vdev);
  if (dev == NULL)
  {
    PDEBUG (2, "dev is NULL, returning");
    return 0;
  }

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
  MOD_INC_USE_COUNT;
#endif
  
  /* lock this device */
  down (&dev->sem);

  /* increment our usage count for the driver */
  ++dev->open;

  /* save our object in the file's private structure */
  file->private_data = vdev;

  zc030x_isoc_init(dev);

  /* unlock this device */
  up (&dev->sem);

  PDEBUG(3,"<< zc030x_open()");
  return retval;
}


/**
  *	zc030x_release
  */
static int
zc030x_release (struct inode *inode, struct file *file)
{
  struct video_device *vdev = video_devdata(file);
  struct usb_zc030x *dev;
  int retval = 0;

  PDEBUG(3,">> zc030x_release()");
  
  if (vdev == NULL)
  {
    PDEBUG (2, "vdev is NULL, returning");
    return 0;
  }

  dev = video_get_drvdata(vdev);
  if (dev == NULL)
  {
    PDEBUG (2, "dev is NULL, returning");
    return 0;
  }
 
  /* lock our device */
  down (&dev->sem);

  if (dev->open <= 0)
    {
      PDEBUG (5, "%s - device not opened", __FUNCTION__);
      retval = -ENODEV;
      //zc030x_unallocate_buffers(dev);
      goto exit_not_opened;
    }
 
  //FIXME dev->streaming = 0;
  /* TODO: what about iso stream? might be still running here*/
  /* might somebody still write in those urbs? lock? */

/*  zc030x_isoc_uninit(dev); 
  zc030x_unallocate_buffers(dev);
*/ 
  if(!dev->present && !dev->open)
  {
      /* the device was unplugged before the file was released */
      up (&dev->sem);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)      
      MOD_DEC_USE_COUNT;
#endif
/*      kfree(dev); */
      return 0;
  }
  
  /* decrement our usage count for the device */
  --dev->open;
  if (dev->open <= 0) {
    dev->open = 0;
  }

  /* decrement our usage count for the module */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
      MOD_DEC_USE_COUNT;
#endif

exit_not_opened:
  up (&dev->sem);
  
  PDEBUG(3,"<< zc030x_release()");
  return retval;
}

//TODO: merged!
/**
 *	zc030x_read
 */
static ssize_t zc030x_read (struct file *file, char *buf, size_t count, loff_t * ppos)
{
  int i;
  int frmx = -1;
  int noblock = file->f_flags&O_NONBLOCK;
  
  struct video_device *vdev = video_devdata(file);
  struct usb_zc030x *dev = video_get_drvdata(vdev);

  volatile struct zc030x_frame *frame;
  
  //return -EFAULT;
  PDEBUG(4, "%d bytes, noblock=%d", count, noblock);

  //return -EFAULT;
  
  /* lock this object */

  if (signal_pending(current))
     return -EINTR;

  if (!dev || !buf)
     return -EFAULT;

  /* lock device object */
  down (&dev->sem);
 
  /* verify that the device wasn't unplugged */
  if (!dev->present) {
      up (&dev->sem);
      return -ENODEV;
  }

   if(!dev->streaming) {
       up(&dev->sem);
      return -EIO;
   }

   PDEBUG(4,"sleep interruptible!");
   interruptible_sleep_on(&dev->wq);
   PDEBUG(4,"signal pending test!");

  if (signal_pending(current)) {
      up( &dev->sem);
      return -EINTR;
  }

  /* 
   * Ok, for the jpeg cams, we know we have received a whole
   * frame now. The rest of the code in this routine confuses 
   * frames and sequence numbers, so we do our own thing here.
   */

  /* One frame has just been set to DONE. Find it. */
  for(i = 0; i < ZC030X_NUMFRAMES; i++)
    if (dev->frame[i].grabstate == FRAME_DONE)
	frmx = i;

  PDEBUG(4, "Frame number: %d", frmx);

  if (frmx < 0)
  {
    /* We havent found a frame that is DONE. Damn. Should
     * not happen. */
    PDEBUG(2, "frame not ready to be read %d", frmx);
    up (&dev->sem);
    return -EFAULT;
  }
  
  frame = &dev->frame[frmx];
  
  // count is ask by users why compute here !!!!
  // best way:
  // if (count > frame->scanlength ) count = frame->scanlength;
  // but frame->scanlength seem to be not used
  // count = frame->width * frame->height * (frame->depth >> 3);
 
  if ((i = copy_to_user(buf, frame->data, count))) 
  {
      PDEBUG(2, "Copy failed! %d bytes not copied", i);
      up (&dev->sem);
      return -EFAULT;
  }
  /* Release the frame */
  frame->grabstate = FRAME_READY;
  /* unlock the device */
  up (&dev->sem);

  return count;
}

static int zc030x_mmap(struct file *file, struct vm_area_struct *vma)

{

  struct video_device *vdev = file->private_data;
  unsigned long start = vma->vm_start;
  unsigned long size  = vma->vm_end - vma->vm_start;
  struct usb_zc030x *dev = video_get_drvdata(vdev);

	unsigned long page, pos;

	if (!dev->present) {
		return -EIO;
	}

	PDEBUG(4, "mmap: %ld (%lX) bytes", size, size);

	if (size > (((ZC030X_NUMFRAMES * MAX_VIDEODATA_SIZE) + PAGE_SIZE - 1) & ~(PAGE_SIZE - 1)))
		return -EINVAL;

	pos = (unsigned long)dev->fbuf;
	while (size > 0) {
		page = kvirt_to_pa(pos);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0) || defined (RH9_REMAP)
		if (remap_page_range(vma, start, page, PAGE_SIZE, PAGE_SHARED))
#else /* RH9_REMAP */
		if (remap_page_range(start, page, PAGE_SIZE, PAGE_SHARED))
#endif /* RH9_REMAP */
			return -EAGAIN;
		start += PAGE_SIZE;
		pos += PAGE_SIZE;
		if (size > PAGE_SIZE)
			size -= PAGE_SIZE;
		else
			size = 0;
	}

	return 0;
}


static int zc030x_find_sensor_by_id(struct usb_device *udev, struct sensor_list * pList, __u32 sizeoflist)
{
    int i             = -1;
    int r             = 0;
    unsigned char tmp = 0; 
    
    // Check arguments
    if (pList == NULL || sizeoflist & 0x80000000) 
        return -1;
    PDEBUG(2, "FSBI : %d", sizeoflist);
    

    // Parse the list
    while (i < (int)sizeoflist-1)
    {
        i++;
        
        // Check if the signature exist
        if (pList[i].i2c_sig == NULL)
            continue;
        
        PDEBUG(2, "FSBI Length : %d", pList[i].i2c_sig->signature_length);
        
        
        // Now read the registers to check if the signature match
        for (r = 0; r < pList[i].i2c_sig->signature_length; r++)
        {
            // Read the specified register
//            zc030x_i2c_reg_write(udev, pList[i].i2c_sig->i2c_reg_addr[r], 0xff);
            tmp = zc030x_i2c_reg_read(udev, pList[i].i2c_sig->i2c_reg_addr[r], 1);
            
            PDEBUG(2, "FSBI [%02X]r(%d) Got: %02X Expected : %02X", pList[i].i2c_sig->i2c_reg_addr[r], r, tmp, pList[i].i2c_sig->i2c_ro_value[r]); 
            // If not match, exit
            if (pList[i].i2c_sig->i2c_ro_value[r] != tmp) 
                break;
        }
        
        // Check if the signature matched
        if (r == pList[i].i2c_sig->signature_length)
            return i;
    }
    
    PDEBUG(2, "FSBI : Not found, sorry.");
    // Not found
    return -1;
}


/**
 *	zc030x_ioctl
 */

//function prototype
static int zc030x_do_ioctl (struct inode *inode, struct file *file, unsigned int cmd, void *arg);

static int zc030x_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	int rc ; 
    PDEBUG (2, ">>> IOCTL arg = %p", arg);
    
    // Zc030x specific ioctl
    switch (cmd)
    {
        case ZCI2CREGISTERMAP:
        {
            unsigned char buf[1];
            unsigned char val[255];
            unsigned char prev = 0;
            int ret;
            int i;
            struct video_device *vdev = file->private_data;
            struct usb_zc030x *dev = video_get_drvdata(vdev);

            struct register_map * prm = arg;
            
            // Write :
            for (i = 0; i < 256; i ++)
            {
                // Save previous value (bug reported on ML)
                prev = zc030x_i2c_reg_read(dev->udev, i, 1);
                prm[i].addr = i;
                prm[i].value = 0xff;
                prm[i].mode = RM_ReadWrite;
                
                zc030x_i2c_reg_write(dev->udev, i, 0xff);
                val[i] = zc030x_i2c_reg_read(dev->udev, i, 1);
                if (val[i] != 0xff)
                {
                    PDEBUG(2, "REGISTERMAP: [FF] (%02X) contains some read only bit with %02X value", i, ret);

                    prm[i].value = val[i];

                    if (val[i] == 0)
                        prm[i].mode = RM_Unused;
                }
                zc030x_i2c_reg_write(dev->udev, i, 0x00);
                ret = zc030x_i2c_reg_read(dev->udev, i, 1);
                if (ret != 0x00)
                {
                    PDEBUG(2, "REGISTERMAP: [00] (%02X) contains some read only bit with %02X value", i, ret);
                }
                if (val[i] == ret && ret != 00)
                {
                    PDEBUG(2, "REGISTERMAP-OK: [ID] (%02X) is read only with %02X value", i, ret);
                    prm[i].addr = i;
                    prm[i].value = ret;
                    prm[i].mode = RM_ReadOnly;
                }
                // Restore previous value
                zc030x_i2c_reg_write(dev->udev, i, prev); 
            }
            rc = 0;
            break;
        }
        case ZCIDSENSOR:
        {
            unsigned char buf[1];
            int ret;
            int i;
            struct video_device *vdev = file->private_data;
            struct usb_zc030x *dev = video_get_drvdata(vdev);
            
            struct sensor_list * psl = arg;
		    PDEBUG (2, ">>> ZCIDSENSOR asked for : %s, (%d)",psl->name, psl->id);
            
            // Let's try to read the i2c first register (I've no idea about how to read an I2C register but...)
            buf[0] = 0x00;

            // Okay, try to find the sensor given its id
            usb_set_interface (dev->udev, 0, 0);
            ret = zc030x_find_sensor_by_id(dev->udev, slist, sizeof(slist) / sizeof(slist[0]));
            usb_set_interface (dev->udev, 0, 7);
            if (ret != -1)
            {
                psl->id = slist[ret].id;
                sprintf(psl->name, "%s", slist[ret].name);
                psl->minwidth = slist[ret].minwidth;
                psl->maxwidth = slist[ret].maxwidth;
                psl->minheight = slist[ret].minheight;
                psl->maxheight = slist[ret].maxheight;
            } else
            {
                psl->id = 0;
                sprintf(psl->name, "Unknown");
                psl->minwidth = 0;
                psl->maxwidth = 0;
                psl->minheight = 0;
                psl->maxheight = 0;
            }
            psl->i2c_sig = NULL;
		    PDEBUG (2, "<<< ZCIDSENSOR asked for : %s, (%d)",psl->name, psl->id);
            rc = 0;
            break;
        }
        default:
            rc = video_usercopy(inode, file, cmd, arg, zc030x_do_ioctl);
            
    }
	// Copy args from user space to kernel space, and call the real ioctl function.
    PDEBUG (2, "<<< IOCTL arg = %p", arg);
	return rc;
}



static int
zc030x_do_ioctl (struct inode *inode, struct file *file, unsigned int cmd, void *arg)
{
     struct video_device *vdev = file->private_data;
     struct usb_zc030x *dev = video_get_drvdata(vdev);
     int ret = -ENOTTY;
     int i;
     int j;
  
     /* lock device object */
     down (&dev->sem);

     /* verify that the device wasn't unplugged */
     if (!dev->present) {
        up (&dev->sem);
        return -ENODEV;
     }

     PDEBUG(2,"zc030x - minor %d, cmd 0x%.4x, arg %ld",dev->minor, cmd, arg);

     /* NOTE: nothing useful done here yet!
      * fill in your device specific stuff here 
      * */
     switch (cmd) 
     {
	     case VIDIOCGCAP:
	     {
		     struct video_capability *b = arg;
		     PDEBUG (2, "VIDIOCGCAP %p :",b);
		     memset(b, 0, sizeof(struct video_capability));
		     i = 0; j = 0;
		     while ((clist[i].id != -1) && (clist[i].id != dev->type_id))
			     i++;
		     while ((slist[j].id != -1) && (slist[j].id != dev->sensor_id))
			     j++;
		     snprintf(b->name,32,"%s with sensor %s\n", clist[i].name, slist[j].name);
		     b->type = VID_TYPE_CAPTURE;
		     b->channels = 1;
		     b->audios = 0;
		     /* for now only 320 x 232
		     b->maxwidth = spca50x->maxwidth;
		     b->maxheight = spca50x->maxheight;
		     b->minwidth = spca50x->minwidth;
		     b->minheight = spca50x->minheight; */
		     b->maxwidth = FrameWidth;
		     b->minwidth = 320;
		     b->maxheight = FrameHeight;
		     b->minheight = 232;
		     /* Value for Webcam Mobile
		     b->minheight = 240;
		     */
		     ret = 0;
		     break;
	     }
	     case VIDIOCGCHAN:
	     {
		     struct video_channel *v = arg;
		     
		     snprintf(v->name,32,"zc030x video bridge\n");
		     v->flags = 0;
		     v->tuners = 0;
		     v->type = VIDEO_TYPE_CAMERA;
		     ret = 0;
		     break;
	     }
	     case VIDIOCSCHAN:
	     {
		     ret = 0;
		     break;
	     }
	     case VIDIOCGPICT:
	     {
		     struct video_picture *p = arg;
		     p->depth = 24;
		     p->palette = VIDEO_PALETTE_RGB24;
		     ret = 0;
		     break;
	     }
	     case VIDIOCGCAPTURE:
	     {
		     int *vf = arg;
		     PDEBUG (4, "VIDIOCGCAPTURE");
		     *vf = 0;
		     // no subcapture
		     ret = -EINVAL;
		     break;
	     }
	     case VIDIOCSCAPTURE:
	     {
		     struct video_capture *vc = arg;
		     /*if (vc->flags)
			     return -EINVAL;
		     if (vc->decimation)
			     return -EINVAL; */
		     ret = -EINVAL;
		     break;
	     }
	     case VIDIOCSWIN:
	     {
		     struct video_window *vw = arg;
		     ret = 0;

		     PDEBUG (3, "VIDIOCSWIN: width=%d, height=%d, flags=%d",
				     vw->width, vw->height,vw->flags);
		     if (vw->x) ret = -EINVAL;
		     if (vw->y) ret = -EINVAL;
		     if (vw->width  != FrameWidth) ret = -EINVAL;
		     if (vw->height != FrameHeight) ret = -EINVAL;
		     break;
	     }
	     case VIDIOCGWIN:
	     {
		     struct video_window *vw = arg;
		     
		     memset(vw, 0, sizeof(struct video_window));
		     vw->x      = 0;
		     vw->y      = 0;
		     vw->width  = FrameWidth;
		     vw->height = FrameHeight;
		     vw->flags  = 0;
		     PDEBUG (4, "VIDIOCGWIN: %dx%d", vw->width, vw->height);
		     
		     ret = 0;
		     break;
	     }
	     case VIDIOCGMBUF:
	     {
		     struct video_mbuf *vm=arg;
		     int i;
		     PDEBUG (2, "VIDIOCGMBUF: %p ", vm);
		     memset(vm, 0, sizeof(struct video_mbuf));
		     vm->size = ZC030X_NUMFRAMES * MAX_VIDEODATA_SIZE;
		     vm->frames = ZC030X_NUMFRAMES;
		     for(i = 0; i < ZC030X_NUMFRAMES; i++)
		     {
			     vm->offsets[i] = MAX_VIDEODATA_SIZE * i;
		     }
		     ret = 0;
		     break;
	     }

	     case VIDIOCSPICT:
	     {
		     struct video_picture *p = arg;

		     ret = (p->palette != VIDEO_PALETTE_RGB24) ? -EINVAL : 0;
		     break;
	     }
	     case VIDIOCMCAPTURE:
	     {
		     struct video_mmap *vm = arg;
		     __u32* buf1;
		     __u32* buf2;
		     PDEBUG(2, "CMCAPTURE, format is %sVIDEO_PALETTE_RGB32", (vm->format == VIDEO_PALETTE_RGB32) ? "" : "not ");
		     PDEBUG(2, "frame: %d, size: %dx%d, format: %d", vm->frame, vm->width, vm->height, vm->format);
		     if ((vm->format != VIDEO_PALETTE_RGB24) && (vm->format != VIDEO_PALETTE_RGB32)) {
			     PDEBUG(1, "VIDIOCMCAPTURE: invalid format (%d)", vm->format);
			     up (&dev->sem);
			     return -EINVAL;
		     }
		     dev->videoformat = vm->format;
		     if ((vm->frame < 0) || (vm->frame > ZC030X_NUMFRAMES)) {
			     PDEBUG(1,"VIDIOCMCAPTURE: invalid frame (%d)", vm->frame);
			     up (&dev->sem);
			     return -EINVAL;
		     }
		     if (vm->width != FrameWidth  || vm->height != FrameHeight) {
			     PDEBUG(1,"VIDIOCMCAPTURE: wrong dimensions. Only %dx%d supported now!", FrameWidth, FrameHeight);
			     up (&dev->sem);
			     return -EINVAL;
		     }
		     /* 
		      * If we are grabbing the current frame, let it pass
		      */ 
		     if (dev->frame[vm->frame].grabstate == FRAME_GRABBING){
			     PDEBUG(4, "MCAPTURE: already grabbing");
			     ret = wait_event_interruptible(dev->wq,
					     (dev->frame[vm->frame].grabstate != FRAME_GRABBING));
			     if (ret){
				     up (&dev->sem);
				     return -EINTR;
			     }
		     }
		     /*
		     //TODO: move this elsewhere!
		     if (dev->videoformat == VIDEO_PALETTE_RGB32)
		     {
			     // if format is RGB32, change from RGB24 to RGB32:
			     for(i = 320 * 232; i > 0; i--){
				     buf1 = (__u32*)((int)(dev->frame[vm->frame].data) + i * 3);
				     buf2 = (__u32*)((int)(dev->frame[vm->frame].data) + i * 4);
				     *buf2 = *buf1;
			     }
		     }
		     */
		     dev->frame[vm->frame].grabstate = FRAME_READY;
		ret = 0;
		break;
	     }
	     case VIDIOCSYNC:
	     {
		     unsigned int frame = *((unsigned int *) arg);
		     __u32* buf1;
		     __u32* buf2;
		     PDEBUG(3, "syncing to frame %d, grabstate = %d", frame,
				     dev->frame[frame].grabstate);
		     switch (dev->frame[frame].grabstate) {
			     case FRAME_UNUSED:
			     {
				     up (&dev->sem);
				     return -EINVAL;
			     }
			     case FRAME_ABORTING:
			     {
				     up (&dev->sem);
				     return -ENODEV;
			     }
			     case FRAME_READY:
			     case FRAME_GRABBING:
			     case FRAME_ERROR:
			     {
redo:
				     if (!dev->present)
				     {
					     up (&dev->sem);
					     return -EIO;
				     }
				     ret = wait_event_interruptible(dev->wq,
						     (dev->frame[frame].grabstate == FRAME_DONE));
				     if (ret)
				     {
					     up (&dev->sem);
					     return -EINTR;
				     }
				     PDEBUG(5,"Synched on frame %d, grabstate = %d",
						     frame, dev->frame[frame].grabstate);
			     
				     if (dev->frame[frame].grabstate == FRAME_ERROR) {
					     goto redo;
				     }
				     /* Fallthrough.
				      * We have waited in state FRAME_GRABBING until it
				      * becomes FRAME_DONE, so now we can move along.
				      */
			     }
			     case FRAME_DONE:
			     {
				     PDEBUG(3,"Already synched %d\n",frame);
				     /* Release the current frame. This means that it
				      * will be reused as soon as all other frames are
				      * full, so the app better be done with it quickly.
				      * Can this be avoided somehow?
				      */
				     //TODO: move this elsewhere !
				     /*
				     if (dev->videoformat == VIDEO_PALETTE_RGB32)
				     {
					     // if format is RGB32, change from RGB24 to RGB32:
					     for(i = 320 * 232; i > 0; i--){
						     buf1 = (__u32*)((int)(dev->frame[frame].data) + i * 3);
						     buf2 = (__u32*)((int)(dev->frame[frame].data) + i * 4);
						     *buf2 = *buf1;
					     }
				     }
				     */

				     dev->frame[frame].grabstate = FRAME_UNUSED;
				     break;
			     }
		     } /* end switch */
		     ret = 0;
		     break;
	     }
	     default:
	     {
		     break;
	     }
     }

     /* unlock the device */
     up (&dev->sem);

     /* return */
     return ret;
}



/**
 *	zc030x_probe
 *
 *	Called by the usb core when a new device is connected that it thinks
 *	this driver might be interested in.
 */

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
static int zc030x_probe(struct usb_interface *intf, const struct usb_device_id *id)
#else
static void *zc030x_probe (struct usb_device *udev, unsigned int ifnum,
	                   const struct usb_device_id *id)
#endif
{
  struct usb_zc030x *dev = NULL;
  struct usb_interface_descriptor *iface;
  struct usb_endpoint_descriptor *endp;
  int i;
  int s_id;
  int actalt;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
  struct usb_device *udev = interface_to_usbdev(intf);
#endif  

  /* We don't handle multi-config cameras - definitely true for wcam300a*/
  if (udev->descriptor.bNumConfigurations != 1)
  {
     PDEBUG(1,"ERROR: Can't handle camera: multiconfiguration device");
     goto nodevice;
  }
  
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
  /* 2.4.x
   * TODO: similar thing for 2.6, see spca50x driver
   * the interfaces are probed one by one. We are only interested in the
   * video interface (0) now.
   * Interface 1 is the Audio Control, and interface 2 Audio itself.
   */
  if (ifnum > 0)
       goto nodevice;
//#endif 
  
  /* Skip probe for alternate settings...*/
  /* alternate setting is fixed right now */
  actalt=udev->actconfig->interface[ifnum].act_altsetting;
#endif //O_o 
  if(actalt > 0)
  {  
     goto nodevice; 
  }
 
  /*** If we reach this point, we found our interface/altsetting */
  PDEBUG(3, ">> zc030x_probe");
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
  MOD_INC_USE_COUNT;
#endif 

  /* allocate memory for our device object and intialize it */
  dev = kmalloc (sizeof (struct usb_zc030x), GFP_KERNEL);
  if (dev == NULL) {
      err ("Couldn't kalloc zc030x device struct!");
      goto error;
  }
  memset (dev, 0x00, sizeof (struct usb_zc030x));
 
  dev->udev = udev;
  
  /* Detect our camera ! */
  if ((zc030x_detect_camera (dev)) < 0)
  {
    PDEBUG(1,"ERROR: Device not supported !");
    goto error;
  }
  
  init_MUTEX (&dev->sem);
  init_MUTEX (&dev->buf_lock);
  
  //usb specific initialization...
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)   
  PDEBUG (3, "Set device configuration to %d\n", DEFAULT_CONFIGURATION);
  usb_set_configuration (udev, DEFAULT_CONFIGURATION);
#endif
  PDEBUG (3, "Set interface to %d\n", VIDEO_INTERFACE);
  PDEBUG (3, "Set alternate configuration %d\n", dev->alt);
  usb_set_interface (udev, VIDEO_INTERFACE, dev->alt);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
  iface = &udev->actconfig->interface[ifnum].altsetting[dev->alt];
#else
 #if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,6)
  iface = &intf->cur_altsetting->desc;
 #else
  iface = &intf->altsetting[dev->alt].desc;
 #endif
#endif
  
  dev->iface_desc = iface;

  /* Endpoint scan procedure was removed, think we can stick with fixed values */
  dev->isocend = 0x01;
  dev->intrend = 0x02;
  
  /* we can register the videodevice now as it is ready... */
  /* register video device (v4l) */
  dev->vdev = video_device_alloc();
  if(!dev->vdev)
    goto error;

  memcpy(dev->vdev, &zc030x_template, sizeof(zc030x_template));
  strcpy(dev->vdev->name, clist[dev->type_id].name);
 
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
  dev->vdev->dev = &udev->dev;
#endif
  video_set_drvdata(dev->vdev, dev);
  
  if (video_register_device(dev->vdev, VFL_TYPE_GRABBER, -1) < 0)
  {
	  err("Unable to register video device");
	  goto error;
  }
  PDEBUG (5, "Registered as video%d.\n",dev->vdev->minor & 0x3F);

  /* TODO: init frame queue - is this done in isoc_init?... */
  for(i = 0; i < ZC030X_NUMFRAMES; i++)
  {
    init_waitqueue_head(&dev->frame[i].wq);
    dev->frame[i].seq=0;
  }
  init_waitqueue_head(&dev->wq);

  /* Configure the zc030x camera */
  if( zc030x_configure_camera(dev) != 0 ) 
  {
     PDEBUG(1, "Error while configuring the camera!");
     goto error;
  }
  
  /* init the jpeg decoder */ 
  s_id = dev->sensor_id;
  /* TODO: pass decoder parameters (image dimensions...) */ 
  InitDecoder();

  /* allow device read,ioctl,mmap */
  dev->present = 1;

  /*FIXME ISOC init moved to open...what about deinit &stuff */
  //zc030x_isoc_init(dev);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
  MOD_DEC_USE_COUNT;
  return dev;
#else
  usb_set_intfdata(intf, dev);
  return 0;
#endif

nodevice:
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
  return NULL; 
#else
  return -ENODEV;
#endif 
  
error:
  /* release video device */
  if(dev->vdev){
    if(dev->vdev->minor == -1)
    video_device_release(dev->vdev);
  else
    video_unregister_device(dev->vdev);
    dev->vdev = NULL;
  }
  if (dev != NULL)
      kfree(dev);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
  MOD_DEC_USE_COUNT;
  return NULL;
#endif  
  return -ENOMEM;
}


/**
 *	zc030x_disconnect
 *
 *	Called by the usb core when the device is removed from the system.
 */
						 
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
static void zc030x_disconnect(struct usb_interface *interface)
{
  struct usb_zc030x *dev =  usb_get_intfdata (interface);
  usb_set_intfdata (interface, NULL);
#else  
static void zc030x_disconnect (struct usb_device *udev, void *ptr)
{
  struct usb_zc030x *dev = (struct usb_zc030x *) ptr;
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
  MOD_INC_USE_COUNT;
#endif

  PDEBUG(3,">> zc030x_disconnect()");
	  
  /* prevent races with open */
  down (&disconnect_sem);

  /* prevent device read, write and ioctl */
  dev->present = 0;

  /* We don't want people trying to open up the device */
  if (dev->vdev) {
    PDEBUG (3, "Unregistered video%d.\n",dev->vdev->minor & 0x3F);
    video_unregister_device(dev->vdev);
    dev->vdev = NULL;
  }
  
  /* if streaming, stop stream and unlink&free urbs */  
  zc030x_isoc_uninit(dev);
  
  /* dellocate buffers */
  zc030x_unallocate_buffers(dev); 
  
  /* deregister intr urb */
  usb_unlink_urb(dev->inturb);
  usb_free_urb(dev->inturb);

  down (&dev->sem);

  /* if the device is not opened, then we clean up right now */
  if (!dev->open) {
    up (&dev->sem);
    kfree (dev);
  }
  else {
    dev->udev = NULL;
    up (&dev->sem);
  }

  up (&disconnect_sem);
 
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
  MOD_DEC_USE_COUNT;
#endif
}


/********************************************************
 *
 * module init & cleanup...
 *
 *********************************************************/

/**
 *	usb_zc030x_init
 */
static int __init
usb_zc030x_init (void)
{
  int result;

  /* Use MODULE_PARM 'debug' for macro PDEBUG */
  if (debug == -1) {
    /* Full debug for developpement if no debug option set */
    pdebug = 5;
  }
  else { pdebug = debug; }

  PDEBUG(1,"Loading experimental zc030x... \n");
  /* register this driver with the USB subsystem */
  result = usb_register (&zc030x_driver);
  if (result < 0)
    {
      err ("usb_register failed for the " __FILE__ " driver. Error number %d",
	   result);
      return -1;
    }

  PDEBUG (5, DRIVER_DESC " " DRIVER_VERSION);

#if defined(CONFIG_PROC_FS) && defined(CONFIG_VIDEO_PROC_FS)
//TODO: proc entry has been temorarily removed  zc030x_proc_create();
#endif

  return 0;
}


/**
 *	usb_zc030x_exit
 */
static void __exit
usb_zc030x_exit (void)
{

#if defined(CONFIG_PROC_FS) && defined(CONFIG_VIDEO_PROC_FS)
//TODO: proc entry management  zc030x_proc_destroy();
#endif
  /* stop isoc stream (if streaming) ) */
    
  /* deregister this driver with the USB subsystem */
  PDEBUG (1,"Unloading experimental zc030x.\n");
  usb_deregister (&zc030x_driver);
}


module_init (usb_zc030x_init);
module_exit (usb_zc030x_exit);

