#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/proc_fs.h>

#include "zc030x_main.h"

#if defined(CONFIG_PROC_FS) && defined(CONFIG_VIDEO_PROC_FS)

static struct proc_dir_entry *zc030x_proc_entry = NULL;
extern struct proc_dir_entry *video_proc_entry;

/**
 * Return some infos about the current cam and the driver
 */
int zc030x_read_proc(char *page, char **start, off_t off, int count, int *eof,
		                     void *data)
{
	char *out = page;
	int len;

	/* IMPORTANT: This output MUST be kept under PAGE_SIZE
	 *            or we need to get more sophisticated. */

	out += sprintf(out, "driver          : %s\n", DRIVER_DESC);
	out += sprintf(out, "driver_version  : %s\n", DRIVER_VERSION);
		
	
	len = out - page;
	len -= off;
	
	if (len < count) {
		*eof = 1;
		if (len <= 0)
			return 0;
	}
	else
		len = count;
	
	*start = page + off;
	
	return len;					
}

/**
 * This fonction create an entry under /proc/video/
 * @todo : include error messages
 */
void zc030x_proc_create(void)
{
	
        if (video_proc_entry == NULL) {
		return;
	}

	zc030x_proc_entry = create_proc_read_entry("zc030x", S_IFREG|S_IRUGO|S_IWUSR,
		                video_proc_entry, zc030x_read_proc, NULL);

	if (zc030x_proc_entry)
		zc030x_proc_entry->owner = THIS_MODULE;
					
}


/**
 * Removes entry in proc
 */
void zc030x_proc_destroy(void)
{
	if (zc030x_proc_entry == NULL)
		return;
	
	remove_proc_entry("zc030x", video_proc_entry);
}
#endif
